@@ -149,7 +149,7 @@ static int if_addressexists(struct interface *, struct in_addr *);
 
 #define PROC_INET6	"/proc/net/if_inet6"
 #define PROC_PROMOTE	"/proc/sys/net/ipv4/conf/%s/promote_secondaries"
-#define SYS_BRIDGE	"/sys/class/net/%s/bridge"
+#define SYS_BRIDGE	"/sys/class/net/%s/bridge/bridge_id"
 #define SYS_LAYER2	"/sys/class/net/%s/device/layer2"
 #define SYS_TUNTAP	"/sys/class/net/%s/tun_flags"
 
@@ -211,7 +211,7 @@ if_machinearch(char *str, size_t len)
 		    fscanf(fp, "%255s", buf) == 1)
 		{
 		        fclose(fp);
-			return snprintf(str, len, ":%s", buf);
+			return snprintf(str, len, "%s", buf);
 		}
 	}
 	fclose(fp);
@@ -220,53 +220,47 @@ if_machinearch(char *str, size_t len)
 }
 
 static int
-check_proc_int(const char *path)
+check_proc_int(struct dhcpcd_ctx *ctx, const char *path)
 {
-	FILE *fp;
-	int i;
+	char buf[64];
+	int error, i;
 
-	fp = fopen(path, "r");
-	if (fp == NULL)
+	if (dhcp_readfile(ctx, path, buf, sizeof(buf)) == -1)
 		return -1;
-	if (fscanf(fp, "%d", &i) != 1)
-		i = -1;
-	fclose(fp);
+	i = (int)strtoi(buf, NULL, 0, INT_MIN, INT_MAX, &error);
+	if (error != 0 && error != ENOTSUP) {
+		errno = error;
+		return -1;
+	}
 	return i;
 }
 
 static int
-check_proc_hex(const char *path, unsigned int *value)
+check_proc_uint(struct dhcpcd_ctx *ctx, const char *path, unsigned int *u)
 {
-	FILE *fp;
-	int i;
+	char buf[64];
+	int error;
 
-	fp = fopen(path, "r");
-	if (fp == NULL)
+	if (dhcp_readfile(ctx, path, buf, sizeof(buf)) == -1)
 		return -1;
-	i = fscanf(fp, "%x", value) == 1 ? 0 : -1;
-	fclose(fp);
-	return i;
+	*u = (unsigned int)strtou(buf, NULL, 0, 0, UINT_MAX, &error);
+	if (error != 0 && error != ENOTSUP) {
+		errno = error;
+		return error;
+	}
+	return 0;
 }
 
 static ssize_t
 if_writepathuint(struct dhcpcd_ctx *ctx, const char *path, unsigned int val)
 {
-	FILE *fp;
-	ssize_t r;
+	char buf[64];
+	int len;
 
-#ifdef PRIVSEP
-	if (ctx->options & DHCPCD_PRIVSEP)
-		return ps_root_writepathuint(ctx, path, val);
-#else
-	UNUSED(ctx);
-#endif
-
-	fp = fopen(path, "w");
-	if (fp == NULL)
+	len = snprintf(buf, sizeof(buf), "%u\n", val);
+	if (len == -1)
 		return -1;
-	r = fprintf(fp, "%u\n", val);
-	fclose(fp);
-	return r;
+	return dhcp_writefile(ctx, path, 0664, buf, (size_t)len);
 }
 
 int
@@ -283,7 +277,7 @@ if_init(struct interface *ifp)
 	 * This matches the behaviour of BSD which makes coding dhcpcd
 	 * a little easier as there's just one behaviour. */
 	snprintf(path, sizeof(path), PROC_PROMOTE, ifp->name);
-	n = check_proc_int(path);
+	n = check_proc_int(ifp->ctx, path);
 	if (n == -1)
 		return errno == ENOENT ? 0 : -1;
 	if (n == 1)
@@ -299,7 +293,7 @@ if_conf(struct interface *ifp)
 
 	/* Some qeth setups require the use of the broadcast flag. */
 	snprintf(path, sizeof(path), SYS_LAYER2, ifp->name);
-	n = check_proc_int(path);
+	n = check_proc_int(ifp->ctx, path);
 	if (n == -1)
 		return errno == ENOENT ? 0 : -1;
 	if (n == 0)
@@ -308,41 +302,40 @@ if_conf(struct interface *ifp)
 }
 
 static bool
-if_bridge(const char *ifname)
+if_bridge(struct dhcpcd_ctx *ctx, const char *ifname)
 {
-	char path[sizeof(SYS_BRIDGE) + IF_NAMESIZE];
-	struct stat sb;
+	char path[sizeof(SYS_BRIDGE) + IF_NAMESIZE], buf[64];
 
 	snprintf(path, sizeof(path), SYS_BRIDGE, ifname);
-	if (stat(path, &sb) == 0 && S_ISDIR(sb.st_mode))
-		return true;
-	return false;
+	if (dhcp_readfile(ctx, path, buf, sizeof(buf)) == -1)
+		return false;
+	return true;
 }
 
 static bool
-if_tap(const char *ifname)
+if_tap(struct dhcpcd_ctx *ctx, const char *ifname)
 {
 	char path[sizeof(SYS_TUNTAP) + IF_NAMESIZE];
-	unsigned int n;
+	unsigned int u;
 
 	snprintf(path, sizeof(path), SYS_TUNTAP, ifname);
-	if (check_proc_hex(path, &n) == -1)
+	if (check_proc_uint(ctx, path, &u) == -1)
 		return false;
-	return n & IFF_TAP;
+	return u & IFF_TAP;
 }
 
 bool
-if_ignore(__unused struct dhcpcd_ctx *ctx, const char *ifname)
+if_ignore(struct dhcpcd_ctx *ctx, const char *ifname)
 {
 
-	if (if_tap(ifname) || if_bridge(ifname))
+	if (if_tap(ctx, ifname) || if_bridge(ctx, ifname))
 		return true;
 	return false;
 }
 
 /* XXX work out Virtal Interface Masters */
 int
-if_vimaster(__unused const struct dhcpcd_ctx *ctx, __unused const char *ifname)
+if_vimaster(__unused struct dhcpcd_ctx *ctx, __unused const char *ifname)
 {
 
 	return 0;
@@ -362,11 +355,12 @@ if_vlanid(const struct interface *ifp)
 }
 
 int
-if_linksocket(struct sockaddr_nl *nl, int protocol)
+if_linksocket(struct sockaddr_nl *nl, int protocol, int flags)
 {
 	int fd;
 
-	if ((fd = xsocket(AF_NETLINK, SOCK_RAW | SOCK_CLOEXEC, protocol)) == -1)
+	fd = xsocket(AF_NETLINK, SOCK_RAW | SOCK_CLOEXEC | flags, protocol);
+	if (fd == -1)
 		return -1;
 	nl->nl_family = AF_NETLINK;
 	if (bind(fd, (struct sockaddr *)nl, sizeof(*nl)) == -1) {
@@ -399,7 +393,7 @@ if_opensockets_os(struct dhcpcd_ctx *ctx)
 	snl.nl_groups |= RTMGRP_IPV6_ROUTE | RTMGRP_IPV6_IFADDR | RTMGRP_NEIGH;
 #endif
 
-	ctx->link_fd = if_linksocket(&snl, NETLINK_ROUTE);
+	ctx->link_fd = if_linksocket(&snl, NETLINK_ROUTE, SOCK_NONBLOCK);
 	if (ctx->link_fd == -1)
 		return -1;
 #ifdef NETLINK_BROADCAST_ERROR
@@ -413,7 +407,7 @@ if_opensockets_os(struct dhcpcd_ctx *ctx)
 
 	ctx->priv = priv;
 	memset(&snl, 0, sizeof(snl));
-	priv->route_fd = if_linksocket(&snl, NETLINK_ROUTE);
+	priv->route_fd = if_linksocket(&snl, NETLINK_ROUTE, 0);
 	if (priv->route_fd == -1)
 		return -1;
 	len = sizeof(snl);
@@ -438,7 +432,7 @@ int
 if_setmac(struct interface *ifp, void *mac, uint8_t maclen)
 {
 	struct ifreq ifr = {
-		.ifr_hwaddr.sa_family = ifp->family,
+		.ifr_hwaddr.sa_family = ifp->hwtype,
 	};
 
 	if (ifp->hwlen != maclen || maclen > sizeof(ifr.ifr_hwaddr.sa_data)) {
@@ -897,9 +891,11 @@ link_netlink(struct dhcpcd_ctx *ctx, void *arg, struct nlmsghdr *nlm)
 	}
 
 	if (nlm->nlmsg_type == RTM_DELLINK) {
+#ifdef PLUGIN_DEV
 		/* If are listening to a dev manager, let that remove
 		 * the interface rather than the kernel. */
 		if (dev_listening(ctx) < 1)
+#endif
 			dhcpcd_handleinterface(ctx, -1, ifn);
 		return 0;
 	}
@@ -916,9 +912,11 @@ link_netlink(struct dhcpcd_ctx *ctx, void *arg, struct nlmsghdr *nlm)
 	/* Check for a new interface */
 	ifp = if_findindex(ctx->ifaces, (unsigned int)ifi->ifi_index);
 	if (ifp == NULL) {
+#ifdef PLUGIN_DEV
 		/* If are listening to a dev manager, let that announce
 		 * the interface rather than the kernel. */
 		if (dev_listening(ctx) < 1)
+#endif
 			dhcpcd_handleinterface(ctx, 1, ifn);
 		return 0;
 	}
@@ -931,12 +929,13 @@ link_netlink(struct dhcpcd_ctx *ctx, void *arg, struct nlmsghdr *nlm)
 	}
 
 	/* Re-read hardware address and friends */
-	if (!(ifi->ifi_flags & IFF_UP) && hwaddr) {
-		uint8_t l;
+	if (!(ifi->ifi_flags & IFF_UP)) {
+		void *hwa = hwaddr != NULL ? RTA_DATA(hwaddr) : NULL;
+		uint8_t hwl = l2addr_len(ifi->ifi_type);
 
-		l = l2addr_len(ifi->ifi_type);
-		if (hwaddr->rta_len == RTA_LENGTH(l))
-			dhcpcd_handlehwaddr(ctx, ifn, RTA_DATA(hwaddr), l);
+		if (hwaddr != NULL && hwaddr->rta_len != RTA_LENGTH(hwl))
+			hwa = NULL;
+		dhcpcd_handlehwaddr(ifp, ifi->ifi_type, hwa, hwl);
 	}
 
 	dhcpcd_handlecarrier(ctx,
@@ -1008,7 +1007,7 @@ if_sendnetlink(struct dhcpcd_ctx *ctx, int protocol, struct nlmsghdr *hdr,
 
 		s = priv->route_fd;
 	} else {
-		if ((s = if_linksocket(&snl, protocol)) == -1)
+		if ((s = if_linksocket(&snl, protocol, 0)) == -1)
 			return -1;
 #ifdef NETLINK_GET_STRICT_CHK
 		if (hdr->nlmsg_type == RTM_GETADDR) {
@@ -1586,11 +1585,12 @@ if_initrt(struct dhcpcd_ctx *ctx, rb_tree_t *kroutes, int af)
 const char *bpf_name = "Packet Socket";
 
 /* Linux is a special snowflake for opening BPF. */
-int
-bpf_open(struct interface *ifp, int (*filter)(struct interface *, int))
+struct bpf *
+bpf_open(const struct interface *ifp,
+    int (*filter)(const struct bpf *, const struct in_addr *),
+    const struct in_addr *ia)
 {
-	struct ipv4_state *state;
-	int s;
+	struct bpf *bpf;
 	union sockunion {
 		struct sockaddr sa;
 		struct sockaddr_ll sll;
@@ -1604,38 +1604,34 @@ bpf_open(struct interface *ifp, int (*filter)(struct interface *, int))
 	int n;
 #endif
 
+	bpf = calloc(1, sizeof(*bpf));
+	if (bpf == NULL)
+		return NULL;
+	bpf->bpf_ifp = ifp;
+
 	/* Allocate a suitably large buffer for a single packet. */
-	state = ipv4_getstate(ifp);
-	if (state == NULL)
-		return -1;
-	if (state->buffer_size < ETH_DATA_LEN) {
-		void *nb;
+	bpf->bpf_size = ETH_DATA_LEN;
+	bpf->bpf_buffer = malloc(bpf->bpf_size);
+	if (bpf->bpf_buffer == NULL)
+		goto eexit;
 
-		if ((nb = realloc(state->buffer, ETH_DATA_LEN)) == NULL)
-			return -1;
-		state->buffer = nb;
-		state->buffer_size = ETH_DATA_LEN;
-		state->buffer_len = state->buffer_pos = 0;
-	}
-
-
-#define SF	SOCK_CLOEXEC | SOCK_NONBLOCK
-	if ((s = xsocket(PF_PACKET, SOCK_RAW | SF, htons(ETH_P_ALL))) == -1)
-		return -1;
-#undef SF
+	bpf->bpf_fd = xsocket(PF_PACKET, SOCK_RAW|SOCK_CXNB,htons(ETH_P_ALL));
+	if (bpf->bpf_fd == -1)
+		goto eexit;
 
 	/* We cannot validate the correct interface,
 	 * so we MUST set this first. */
-	if (bind(s, &su.sa, sizeof(su.sll)) == -1)
+	if (bind(bpf->bpf_fd, &su.sa, sizeof(su.sll)) == -1)
 		goto eexit;
 
-	if (filter(ifp, s) != 0)
+	if (filter(bpf, ia) != 0)
 		goto eexit;
 
 	/* In the ideal world, this would be set before the bind and filter. */
 #ifdef PACKET_AUXDATA
 	n = 1;
-	if (setsockopt(s, SOL_PACKET, PACKET_AUXDATA, &n, sizeof(n)) != 0) {
+	if (setsockopt(bpf->bpf_fd, SOL_PACKET, PACKET_AUXDATA,
+	    &n, sizeof(n)) != 0) {
 		if (errno != ENOPROTOOPT)
 			goto eexit;
 	}
@@ -1650,51 +1646,54 @@ bpf_open(struct interface *ifp, int (*filter)(struct interface *, int))
 	 * Or to put it another way, don't trust the Linux BPF filter.
 	*/
 
-	return s;
+	return bpf;
 
 eexit:
-	close(s);
-	return -1;
+	if (bpf->bpf_fd != -1)
+		close(bpf->bpf_fd);
+	free(bpf->bpf_buffer);
+	free(bpf);
+	return NULL;
 }
 
 /* BPF requires that we read the entire buffer.
  * So we pass the buffer in the API so we can loop on >1 packet. */
 ssize_t
-bpf_read(struct interface *ifp, int s, void *data, size_t len,
-    unsigned int *flags)
+bpf_read(struct bpf *bpf, void *data, size_t len)
 {
 	ssize_t bytes;
-	struct ipv4_state *state = IPV4_STATE(ifp);
-
 	struct iovec iov = {
-		.iov_base = state->buffer,
-		.iov_len = state->buffer_size
+		.iov_base = bpf->bpf_buffer,
+		.iov_len = bpf->bpf_size,
 	};
 	struct msghdr msg = { .msg_iov = &iov, .msg_iovlen = 1 };
 #ifdef PACKET_AUXDATA
-	unsigned char cmsgbuf[CMSG_LEN(sizeof(struct tpacket_auxdata))] = { 0 };
+	union {
+		struct cmsghdr hdr;
+		uint8_t buf[CMSG_SPACE(sizeof(struct tpacket_auxdata))];
+	} cmsgbuf = { .buf = { 0 } };
 	struct cmsghdr *cmsg;
 	struct tpacket_auxdata *aux;
 #endif
 
 #ifdef PACKET_AUXDATA
-	msg.msg_control = cmsgbuf;
-	msg.msg_controllen = sizeof(cmsgbuf);
+	msg.msg_control = cmsgbuf.buf;
+	msg.msg_controllen = sizeof(cmsgbuf.buf);
 #endif
 
-	bytes = recvmsg(s, &msg, 0);
+	bytes = recvmsg(bpf->bpf_fd, &msg, 0);
 	if (bytes == -1)
 		return -1;
-	*flags |= BPF_EOF; /* We only ever read one packet. */
-	*flags &= ~BPF_PARTIALCSUM;
+	bpf->bpf_flags |= BPF_EOF; /* We only ever read one packet. */
+	bpf->bpf_flags &= ~BPF_PARTIALCSUM;
 	if (bytes) {
-		if (bpf_frame_bcast(ifp, state->buffer) == 0)
-			*flags |= BPF_BCAST;
+		if (bpf_frame_bcast(bpf->bpf_ifp, bpf->bpf_buffer) == 0)
+			bpf->bpf_flags |= BPF_BCAST;
 		else
-			*flags &= ~BPF_BCAST;
+			bpf->bpf_flags &= ~BPF_BCAST;
 		if ((size_t)bytes > len)
 			bytes = (ssize_t)len;
-		memcpy(data, state->buffer, (size_t)bytes);
+		memcpy(data, bpf->bpf_buffer, (size_t)bytes);
 #ifdef PACKET_AUXDATA
 		for (cmsg = CMSG_FIRSTHDR(&msg);
 		     cmsg;
@@ -1704,7 +1703,7 @@ bpf_read(struct interface *ifp, int s, void *data, size_t len,
 			    cmsg->cmsg_type == PACKET_AUXDATA) {
 				aux = (void *)CMSG_DATA(cmsg);
 				if (aux->tp_status & TP_STATUS_CSUMNOTREADY)
-					*flags |= BPF_PARTIALCSUM;
+					bpf->bpf_flags |= BPF_PARTIALCSUM;
 			}
 		}
 #endif
@@ -1721,7 +1720,17 @@ bpf_attach(int s, void *filter, unsigned int filter_len)
 	};
 
 	/* Install the filter. */
-	return setsockopt(s, SOL_SOCKET, SO_ATTACH_FILTER, &pf, sizeof(pf));
+	if (setsockopt(s, SOL_SOCKET, SO_ATTACH_FILTER, &pf, sizeof(pf)) == -1)
+		return -1;
+
+#ifdef SO_LOCK_FILTER
+	int on = 1;
+
+	if (setsockopt(s, SOL_SOCKET, SO_LOCK_FILTER, &on, sizeof(on)) == -1)
+		return -1;
+#endif
+
+	return 0;
 }
 
 int
@@ -1855,14 +1864,19 @@ int
 if_addrflags6(const struct interface *ifp, const struct in6_addr *addr,
     __unused const char *alias)
 {
-	FILE *fp;
+	char buf[PS_BUFLEN], *bp = buf, *line;
+	ssize_t buflen;
 	char *p, ifaddress[33], address[33], name[IF_NAMESIZE + 1];
 	unsigned int ifindex;
 	int prefix, scope, flags, i;
 
-	fp = fopen(PROC_INET6, "r");
-	if (fp == NULL)
+	buflen = dhcp_readfile(ifp->ctx, PROC_INET6, buf, sizeof(buf));
+	if (buflen == -1)
 		return -1;
+	if ((size_t)buflen == sizeof(buf)) {
+		errno = ENOBUFS;
+		return -1;
+	}
 
 	p = ifaddress;
 	for (i = 0; i < (int)sizeof(addr->s6_addr); i++) {
@@ -1870,23 +1884,20 @@ if_addrflags6(const struct interface *ifp, const struct in6_addr *addr,
 	}
 	*p = '\0';
 
-	while (fscanf(fp, "%32[a-f0-9] %x %x %x %x %"TOSTRING(IF_NAMESIZE)"s\n",
-	    address, &ifindex, &prefix, &scope, &flags, name) == 6)
-	{
-		if (strlen(address) != 32) {
-			fclose(fp);
+	while ((line = get_line(&bp, &buflen)) != NULL) {
+		if (sscanf(line,
+		    "%32[a-f0-9] %x %x %x %x %"TOSTRING(IF_NAMESIZE)"s\n",
+		    address, &ifindex, &prefix, &scope, &flags, name) != 6 ||
+		    strlen(address) != 32)
+		{
 			errno = EINVAL;
 			return -1;
 		}
 		if (strcmp(name, ifp->name) == 0 &&
 		    strcmp(ifaddress, address) == 0)
-		{
-			fclose(fp);
 			return flags;
-		}
 	}
 
-	fclose(fp);
 	errno = ESRCH;
 	return -1;
 }
@@ -1961,11 +1972,12 @@ static const char *p_neigh = "/proc/sys/net/ipv6/neigh";
 void
 if_setup_inet6(const struct interface *ifp)
 {
+	struct dhcpcd_ctx *ctx = ifp->ctx;
 	int ra;
 	char path[256];
 
 	/* The kernel cannot make stable private addresses. */
-	if (if_disable_autolinklocal(ifp->ctx, ifp->index) == -1)
+	if (if_disable_autolinklocal(ctx, ifp->index) == -1 && errno != ENODEV)
 		logdebug("%s: if_disable_autolinklocal", ifp->name);
 
 	/*
@@ -1976,21 +1988,21 @@ if_setup_inet6(const struct interface *ifp)
 		return;
 
 	snprintf(path, sizeof(path), "%s/%s/autoconf", p_conf, ifp->name);
-	ra = check_proc_int(path);
+	ra = check_proc_int(ctx, path);
 	if (ra != 1 && ra != -1) {
-		if (if_writepathuint(ifp->ctx, path, 0) == -1)
+		if (if_writepathuint(ctx, path, 0) == -1)
 			logerr("%s: %s", __func__, path);
 	}
 
 	snprintf(path, sizeof(path), "%s/%s/accept_ra", p_conf, ifp->name);
-	ra = check_proc_int(path);
+	ra = check_proc_int(ctx, path);
 	if (ra == -1) {
 		/* The sysctl probably doesn't exist, but this isn't an
 		 * error as such so just log it and continue */
 		if (errno != ENOENT)
 			logerr("%s: %s", __func__, path);
 	} else if (ra != 0) {
-		if (if_writepathuint(ifp->ctx, path, 0) == -1)
+		if (if_writepathuint(ctx, path, 0) == -1)
 			logerr("%s: %s", __func__, path);
 	}
 }
@@ -2003,18 +2015,25 @@ if_applyra(const struct ra *rap)
 	struct dhcpcd_ctx *ctx = rap->iface->ctx;
 	int error = 0;
 
-	snprintf(path, sizeof(path), "%s/%s/hop_limit", p_conf, ifname);
-	if (if_writepathuint(ctx, path, rap->hoplimit) == -1)
-		error = -1;
+	if (rap->hoplimit != 0) {
+		snprintf(path, sizeof(path), "%s/%s/hop_limit", p_conf, ifname);
+		if (if_writepathuint(ctx, path, rap->hoplimit) == -1)
+			error = -1;
+	}
 
-	snprintf(path, sizeof(path), "%s/%s/retrans_time_ms", p_neigh, ifname);
-	if (if_writepathuint(ctx, path, rap->retrans) == -1)
-		error = -1;
+	if (rap->retrans != 0) {
+		snprintf(path, sizeof(path), "%s/%s/retrans_time_ms",
+		    p_neigh, ifname);
+		if (if_writepathuint(ctx, path, rap->retrans) == -1)
+			error = -1;
+	}
 
-	snprintf(path, sizeof(path), "%s/%s/base_reachable_time_ms",
-	    p_neigh, ifname);
-	if (if_writepathuint(ctx, path, rap->reachable) == -1)
-		error = -1;
+	if (rap->reachable != 0) {
+		snprintf(path, sizeof(path), "%s/%s/base_reachable_time_ms",
+		    p_neigh, ifname);
+		if (if_writepathuint(ctx, path, rap->reachable) == -1)
+			error = -1;
+	}
 
 	return error;
 }
@@ -2022,14 +2041,18 @@ if_applyra(const struct ra *rap)
 int
 ip6_forwarding(const char *ifname)
 {
-	char path[256];
-	int val;
+	char path[256], buf[64];
+	int error, i;
 
 	if (ifname == NULL)
 		ifname = "all";
 	snprintf(path, sizeof(path), "%s/%s/forwarding", p_conf, ifname);
-	val = check_proc_int(path);
-	return val == -1 ? 0 : val;
+	if (readfile(path, buf, sizeof(buf)) == -1)
+		return 0;
+	i = (int)strtoi(buf, NULL, 0, INT_MIN, INT_MAX, &error);
+	if (error != 0)
+		return 0;
+	return i;
 }
 
 #endif /* INET6 */
