@@ -104,7 +104,6 @@ logprintdate(FILE *stream)
 		return -1;
 
 	now = tv.tv_sec;
-	tzset();
 	if (localtime_r(&now, &tmnow) == NULL)
 		return -1;
 	if (strftime(buf, sizeof(buf), "%b %d %T ", &tmnow) == 0)
@@ -238,6 +237,7 @@ vlogerrmessage(int pri, const char *fmt, va_list args)
 
 	vsnprintf(buf, sizeof(buf), fmt, args);
 	logmessage(pri, "%s: %s", buf, strerror(_errno));
+	errno = _errno;
 }
 
 __printflike(2, 3) void
@@ -366,6 +366,9 @@ logopen(const char *path)
 {
 	struct logctx *ctx = &_logctx;
 
+	/* Cache timezone */
+	tzset();
+
 	if (path == NULL) {
 		int opts = 0;
 
