@@ -60,6 +60,13 @@
 # endif
 #endif
 
+#ifdef __linux__
+# include <linux/version.h> /* RTA_PREF is only an enum.... */
+# if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 1, 0)
+#  define HAVE_ROUTE_PREF
+# endif
+#endif
+
 #if defined(__OpenBSD__) || defined (__sun)
 #  define ROUTE_PER_GATEWAY
 /* XXX dhcpcd doesn't really support this yet.
@@ -86,6 +93,14 @@ struct rt {
 #ifdef HAVE_ROUTE_METRIC
 	unsigned int		rt_metric;
 #endif
+#ifdef HAVE_ROUTE_PREF
+	int			rt_pref;
+#endif
+#define RTPREF_HIGH	1
+#define RTPREF_MEDIUM	0	/* has to be zero */
+#define RTPREF_LOW	(-1)
+#define RTPREF_RESERVED	(-2)
+#define RTPREF_INVALID	(-3)	/* internal */
 	unsigned int		rt_dflags;
 #define	RTDF_IFA_ROUTE		0x02		/* Address generated route */
 #define	RTDF_FAKE		0x04		/* Maybe us on lease reboot  */
